###############################################################################
# Author	: Gilles Darold <gilles@darold.net>
# Status	: production / stable
# Purpose	:
#
#    LDAP Single Signon User Authentication with cache based on cookie
#	!!! Apache2 / mod_perl2 version only !!!
#
# Copyright (C) 2002-2005, Gilles Darold - Groupe SAMSE
# All Rights Reserved.
# 
# This module is free software; you can redistribute it and/or
# modify it under the terms of Perl itself.
#
###############################################################################


# Package name
package Apache2::AuthenCacheLDAP;


# Required libraries
use strict;
use mod_perl2();
use Apache2::Access();
use Apache2::Connection();
use Apache2::Log();
use Apache2::RequestRec();
use Apache2::RequestIO();
use Apache2::RequestUtil();
use APR::Table();
use Apache2::Const -compile => qw(AUTH_REQUIRED OK SERVER_ERROR);

use Net::LDAP qw(:all);
use MIME::Base64;

# Global variables
$Apache2::AuthenCacheLDAP::VERSION = '2.0';


###############################################################################
# handler: hook into Apache/mod_perl API
###############################################################################
sub handler
{
	my $r = shift;

	return Apache2::Const::OK unless $r->is_initial_req; # only the first internal request

	# Get all configuration option
	my $cache_auth = $r->dir_config('AuthCache') || 'off';
	my $cache_noport = $r->dir_config('AuthCacheSuppressPort') || 'off';
	my $cache_sso = $r->dir_config('AuthCacheSingleSignon') || 'off';
	my $cache_ttl = $r->dir_config('AuthCacheTTL') || 0;
	my $basedn = $r->dir_config('BaseDN') || '';
	my $ldapsrv = $r->dir_config('LDAPServer') || 'localhost';
	my @ldapserver = ();
	push(@ldapserver, split(/[;,]+/, $ldapsrv));
	my $ldapport = $r->dir_config('LDAPPort') || 389;
	my $uidattr = $r->dir_config('UIDAttr') || 'uid';
	my $basegrpdn = $r->dir_config('BaseGroupDN') || '';
	my $grpattr = $r->dir_config('GRPAttr') || 'memberUid';
	my $linkattr = $r->dir_config('LinkAttr') || '';
	my $extra = $r->dir_config('ExtraAttr') || '';
	my @extras = split(/[;,]+/, $extra);
	my $logout_str = $r->dir_config('LogoutPattern') || 'logout';
	$logout_str = '' if ((lc($logout_str) eq 'off') || !$logout_str);
	# Get requirement options
	my $requires = $r->requires;  
	return Apache2::Const::OK unless $requires; # No requirement mean no auth
	# Enable cache if require
	if ((lc($cache_auth) eq 'on') || ($cache_auth eq '1')) {
		# The coockie name for AuthCacheLDAP is construct as follow :
		# Apache-AuthCache-
		# Domaine name (.samse.fr)
		# :Port (:80)
		# -Auth_Type
		# :Auth_name
		# =value;
		my $server_name = $r->get_server_name() || '';
		my $domain_name = $server_name || '';
		if ((lc($cache_sso) eq 'on') || ($cache_sso eq '1')) { 
			$domain_name =~ s/^[^\.]+//;
		}
		my $server_port = $r->get_server_port() || '';
		$server_port = 0 if ((lc($cache_noport) eq 'on') || ($cache_noport eq '1'));
		my $auth_type = $r->auth_type() || '';
		my $auth_name = $r->auth_name() || '';
		my @coockies = ();
		push(@coockies, split(/;/, $r->headers_in->get('Cookie') || ''));
		my $value = '';
		foreach my $cookie (@coockies) {
			$cookie =~ s/^\s+//;
			if ($cookie =~ /^Apache-AuthCache-$domain_name:$server_port-$auth_type:$auth_name=(.*)/) {
				$value = decode_base64($1);
			}
		}
		if ($value) {
			my $remote_ip = $r->connection->remote_ip() || '';
			# value => username:passwd:remote_ip:timestamp:remote_group:remote_allgroup:other_name=other_val";
			my ($us,$pw,$ip,$tm,$rg,$rag,@others) = split(/:/, $value);
			my $name = $r->user || '';
			$pw = decode_base64($pw);
			# Check the ip address of the host
			if (($us eq $name) || ($us && ($name eq '')) && ($remote_ip eq $ip)) {
				# Check if the requirement are the same or at least that the requiered group is found
				my @allowedgrp = ();
				for my $req (@$requires) {
					my ($require, $rest) = split /\s+/, $req->{requirement}, 2;
					if ($require eq 'user') {
						if ( !grep($us eq $_, split(/\s+/, $rest)) ) {
						    $r->note_basic_auth_failure;
						    $r->log_reason("user $us not allowed", $r->uri);
						    return Apache2::Const::AUTH_REQUIRED;
						}
					}
					if ( ($require eq 'group') || ($require eq 'ldap-group') ) {
						my @allowedgrp = split(/\s+/, $rest);
						if ($#allowedgrp >= 0) {
							if (!$rag && !$rg) {
								$r->note_basic_auth_failure;
								$r->log_reason("user $us: have no group set",$r->uri);
								return Apache2::Const::AUTH_REQUIRED;
							}
							my $groups = ';' . $rag;
							my $found = 0;
							foreach my $g (@allowedgrp) {
								if ($groups =~ /;$g;/) {
									$found = 1;
									$rg = $g;
									last;
								}
							}
							if (!$found) {
								$r->note_basic_auth_failure;
								$r->log_reason("user $us: doesn't match allowed group",$r->uri);
								return Apache2::Const::AUTH_REQUIRED;
							}
						}
					}
				}
				# check cache expiration
				my $expired = 0;
				if ($tm && $cache_ttl) {
					my $current = time;
					if ($current > ($tm + $cache_ttl)) {
						$expired = 1;
					}
				}
				my $args = $r->args;
				if ( $logout_str && ($args =~ /$logout_str=(\d+)/) ) {
					if ($r->request_time < $1) {
						$r->user('');
						$r->subprocess_env(REMOTE_USER => '');
						$r->subprocess_env(REMOTE_GROUP => '');
						$r->subprocess_env(REMOTE_ALLGROUP => '');
						foreach my $o (@others) {
							my ($nm, $val) = split(/=/, $o);
							$nm = 'REMOTE_' . uc($nm);
							$r->subprocess_env($nm => '');
						}
						# Not sure the one is working better
						$r->headers_out->set('Authorization' => '');
						$r->headers_out->set('Set-Cookie' => "Apache-AuthCache-$domain_name:$server_port-$auth_type:$auth_name=; expires=" . scalar gmtime($r->request_time - 86400). "; path=/; domain=$domain_name;");
						$r->err_headers_out->set('Authorization' => '');
						$r->err_headers_out->set('Set-Cookie' => "Apache-AuthCache-$domain_name:$server_port-$auth_type:$auth_name=; expires=" . scalar gmtime($r->request_time - 86400). "; path=/; domain=$domain_name;");
						$r->note_basic_auth_failure;
						$r->log_error("user $us: request for logout");
						$r->status(401);
						return 0;
					}
				}
				if (!$expired) {
					# Restore authorization
					my $auth = encode_base64("$us:$pw") || '';
					$r->headers_in->set('Authorization' => $auth);
					# set authorization type from cookie value
					$r->auth_type($auth_type);
					# Set all environment variable
					$r->user($us);
					$r->subprocess_env(REMOTE_GROUP => $rg);
					$r->subprocess_env(REMOTE_ALLGROUP => $rag);
					foreach my $o (@others) {
						my ($nm, $val) = split(/=/, $o);
						$nm = 'REMOTE_' . uc($nm);
						$r->subprocess_env($nm => $val);
					}
					#$r->log_error("Authentication retrieved from cookie cache for user $us - group: $rg - allgroup: $rag");
					return Apache2::Const::OK;
				} else {
					$r->log_reason("user $us: cache has expired",$r->uri);
				}
			} elsif ($remote_ip ne $ip) {
				$r->note_basic_auth_failure;
				$r->log_reason("address $ip for user $us: doesn't match remote address",$r->uri);
				return Apache2::Const::AUTH_REQUIRED;
			}
		}
	}

	# Get password sent
	my ($res, $sent_pwd) = $r->get_basic_auth_pw;
	return $res if $res;

	# Get user sent
	my $name = $r->user || '';
	unless ($name) {
		$r->note_basic_auth_failure;
		$r->log_reason("no username supplied", $r->uri);
		return Apache2::Const::AUTH_REQUIRED;
	}
  
	#$r->log_error("Passing throught normal authentication for user $name");

	my @allowedgrp = ();
	for my $req (@$requires) {
		my ($require, $rest) = split /\s+/, $req->{requirement}, 2;
		if ($require eq 'user') {
			if ( !grep($name eq $_, split(/\s+/, $rest)) ) {
			    $r->note_basic_auth_failure;
			    $r->log_reason("user $name not allowed", $r->uri);
			    return Apache2::Const::AUTH_REQUIRED;
			}
		}
		if ($require eq 'group') {
			@allowedgrp = split(/\s+/, $rest);
		}
	}

	if ($sent_pwd eq '') {
		$r->note_basic_auth_failure;
		$r->log_reason("user $name: no password supplied", $r->uri);
		return Apache2::Const::AUTH_REQUIRED;
	}
  
	# Connect to the server
	my $ld = undef;
	foreach my $ldap (@ldapserver) {
		$ld = new Net::LDAP($ldap, port => $ldapport);
		last if (defined $ld);
	}
	if (!defined $ld) {
		$r->note_basic_auth_failure;
		$r->log_reason("user $name: LDAP Connection Failed", $r->uri);
		return Apache2::Const::SERVER_ERROR;
	}

	# Bind anonymously
	my $msg = $ld->bind;
	unless ($msg->code == LDAP_SUCCESS) {
		$r->note_basic_auth_failure;
		$r->log_reason("user $name: LDAP Initial Bind Failed: " . $msg->code . " " . $msg->error, $r->uri);
		return Apache2::Const::SERVER_ERROR;
	}
  
	# Create the filter and search
	my $filter = "($uidattr=$name)";
	$r->log->debug("Using filter: $filter");
	my $attrs = [$uidattr];
	push(@$attrs, $linkattr) if ($linkattr);
	push(@$attrs, @extras) if ($#extras >= 0);

	$msg = $ld->search(
		base => $basedn,
		filter => $filter,
		attrs => $attrs
	);
	unless ($msg->code == LDAP_SUCCESS) {
		$r->note_basic_auth_failure;
		$r->log_reason("user $name: ldap search operation failed: " . $msg->code . " " . $msg->error, $r->uri);
		return Apache2::Const::SERVER_ERROR;
	}

	# Did we receive any entries
	unless ($msg->count) {
		$r->note_basic_auth_failure;
		$r->log_reason("user $name: username not found",$r->uri);
		return Apache2::Const::AUTH_REQUIRED;
	}
  
	# Only want the first to get the DN of the user
	my $entry = $msg->first_entry;
	my $dn = $entry->dn;

	# Set the environment variable for each required extra attributes
	my $max = $msg->count;
	my %extra_attr = ();
	for (my $i = 0 ; $i < $max ; $i++) {
		my $entry = $msg->entry($i);
		foreach my $attr ($entry->attributes) {
			next if ($attr eq $uidattr);
			my $val = $entry->get_value($attr);
			$extra_attr{$attr} = $val if (grep(/^$attr$/i, @extras));
		}
	}
	my $others = '';
	foreach my $key (keys %extra_attr) {
		my $name = "REMOTE_" . uc($key);
		$others .= ":$key=" . $extra_attr{$key};
		$r->subprocess_env($name => $extra_attr{$key});
	}

	# Bind as the user we're authenticating
	$msg = $ld->bind($dn, password => $sent_pwd);
	unless ($msg->code == LDAP_SUCCESS) {
		$r->note_basic_auth_failure;
		$r->log_reason("user $name: password mismatch", $r->uri);
		return Apache2::Const::AUTH_REQUIRED;
	}

	# Look for user group based on GroupAttr
	my $groups = '';
	if ($basegrpdn) {
		my $grpmemberuid = $name;
		$grpmemberuid = $extra_attr{$linkattr} if ($linkattr);
		$attrs = [ 'cn' ];
		$msg = $ld->search(
			base => $basegrpdn,
			scope => 'sub',
			filter => "($grpattr=$grpmemberuid)",
			attrs => $attrs
		);

		unless ($msg->code == LDAP_SUCCESS) {
			$r->note_basic_auth_failure;
			$r->log_reason("user $name: ldap search operation failed: " . $msg->code . " " . $msg->error, $r->uri);
			return Apache2::Const::SERVER_ERROR;
		}

		# Did we receive any entries
		if ($msg->count) {
			# Retrieved all groups
			my $max = $msg->count;
			if ($max) {
				for(my $i = 0 ; $i < $max ; $i++) {
					my $entry = $msg->entry($i);
					$groups .= $entry->get_value('cn') . ";";
				}
			}
			# Set the environment variable
			$r->subprocess_env(REMOTE_ALLGROUP => $groups);
		}
	}
	$ld->unbind;

	my $remote_group = '';
	if ($#allowedgrp >= 0) {
		if (!$groups) {
			$r->note_basic_auth_failure;
			$r->log_reason("user $name: have no group set",$r->uri);
			return Apache2::Const::AUTH_REQUIRED;
		}
		$groups = ';' . $groups;
		my $found = 0;
		foreach my $g (@allowedgrp) {
			if ($groups =~ /;$g;/) {
				$found = 1;
				$r->subprocess_env(REMOTE_GROUP => $g);
				$remote_group = $g;
				last;
			}
		}
		if (!$found) {
			$r->note_basic_auth_failure;
			$r->log_reason("user $name: doesn't match allowed group",$r->uri);
			return Apache2::Const::AUTH_REQUIRED;
		}
	}

	if ((lc($cache_auth) eq 'on') || ($cache_auth eq '1')) {
		# The coockie name for AuthCacheLDAP is construct as follow :
		# Apache-AuthCache-
		# Domaine name (.samse.fr)
		# :Port (:80)
		# -Auth_Type
		# :Auth_name
		# =value;
		my $server_name = $r->get_server_name() || '';
		my $domain_name = $server_name || '';
		if ((lc($cache_sso) eq 'on') || ($cache_sso eq '1')) { 
			$domain_name =~ s/^[^\.]+//;
		}
		my $server_port = $r->get_server_port();
		$server_port = 0 if ((lc($cache_noport) eq 'on') || ($cache_noport eq '1'));
		my $auth_type = $r->auth_type();
		my $auth_name = $r->auth_name();
		my $pwd = encode_base64($sent_pwd, '');
		my $ip = $r->connection->remote_ip();
		my $tm = time;
		#  value = username:passwd:remote_ip:timestamp:remote_group:remote_allgroup:other_name=other_val";
		my $value = "$name:$pwd:$ip:$tm:$remote_group:$groups$others";
		$value = encode_base64($value, '');
		$r->headers_out->set('Set-Cookie' => "Apache-AuthCache-$domain_name:$server_port-$auth_type:$auth_name=$value; path=/; domain=$domain_name;");
	}

	# Everything's A-OK
	return Apache2::Const::OK;
}

1;

__END__

# Documentation - try 'pod2text AuthenCacheLDAP'

=head1 NAME

Apache2::AuthenCacheLDAP - mod_perl2 LDAP Authentication Module

=head1 SYNOPSIS

Here is a simple authentication configuration. No single signon
and user simple authentication:

    <Directory "/opt/foo">
        AllowOverride All
        Options ExecCGI FollowSymLinks
        # Authentication Realm and Type (only Basic supported)
        AuthName "Foo Authentication"
        AuthType Basic
        PerlSetVar BaseDN "ou=people,dc=foo,dc=com"
        PerlSetVar LDAPServer "ldap.foo.com"
        PerlSetVar LDAPPort "389"
        PerlSetVar UIDAttr "uid"
        PerlAuthenHandler Apache2::AuthenCacheLDAP
        require valid-user
        Order allow,deny
        Allow from all
    </Directory>


Here the more complexe authentication setting you can do with this module
with single signon (cookie caching) and user/group authentication:

    <Directory "/opt/foo">
        AllowOverride All
        Options ExecCGI FollowSymLinks
        # Authentication Realm and Type (only Basic supported)
        AuthName "Foo Authentication"
        AuthType Basic
        AuthGroupFile /dev/null
        PerlSetVar AuthCache "on"
        PerlSetVar AuthCacheSuppressPort "on"
        PerlSetVar BaseDN "ou=people,dc=foo,dc=com"
        PerlSetVar LDAPServer "ldap.foo.com"
        PerlSetVar LDAPPort "389"
        PerlSetVar UIDAttr "uid"
        PerlSetVar UserGroupDN "ou=group,dc=foo,dc=com" 
        PerlSetVar GRPAttr "memberUid"
        PerlSetVar LinkAttr "employeeNumber"
        PerlSetVar ExtraAttr "cn;departmentNumber;mail"
        PerlSetVar AllowGroup "intranet"
        PerlAuthenHandler Apache2::AuthenCacheLDAP
        require ldap-group support_info
        require valid-user
        satisfy all
        Order allow,deny
        Allow from all
    </Directory>

It also set and restore three LDAP based variable: cn, departmentNumber, mail
that you can acces with $ENV{REMOTE_CN}, $ENV{REMOTE_DEPARTMENTNUMBER} and $ENV{REMOTE_MAIL}

=head1 DESCRIPTION

B<Apache2::AuthenCacheLDAP> is designed to work with mod_perl2 and Net::LDAP.
This module authenticates a user against an LDAP backend.

The authentication process is as follow:

1) Connect to the LDAP server.

2) Check if the given user is allowed if require user is set.

3) Authenticate with the given login/password.

4) Authenticate with group if BaseGroupDN and AllowGroup is set.

Additionnaly this module allow single signon based on a domain and
caching by using cookie. It also allow setting of Apache environment
variable based on the LDAP attribute.


=head1 AUTHENTICATION CONFIGURATION OPTIONS

The following variables can be defined within the configuration
of Directory, Location, or Files blocks or within .htaccess
files.

=over 4

=item B<BaseDN>

The base distinguished name with which to query LDAP. By default,
the BaseDN is empty.

=back

=over 4

=item B<LDAPServer>

The hostname for the LDAP server to query. By default,
LDAPServer is set to localhost. To allow failover, you
can pass a list of LDAP server separated by semi colon

example : master.domain.com;slave1.domain.com;slave2.domain.com

When the perl module try to connect to a server and the connection
fail, it goes to the second server, etc.

=back

=over 4

=item B<LDAPPort>

The port on which the LDAP server is listening. By default,
AuthenCacheLDAPPort is set to 389.

=back

=over 4

=item B<UidAttr>

The attribute name that contains the user's
identification. By default, UidAttr is set to uid.

=back

=over 4

=item B<BaseGroupDN>

The base distinguished name used to query a LDAP posix group.
By default, the BaseGroupDN is empty.

=back
 
=over 4

=item B<GRPAttr>

The attribute name that contains the user's identification into the
posix group object. By default, GRPAttr is set to memberUid. And the
value is retrieved from the UidAttr value.

=back
 
=over 4

=item B<LinkAttr>

The attribute name that contains the user's identification to search
into the posix group object as value of the GRPAttr. See above.
By default, LinkAttr is set to the UidAttr.

This is to use if the user DN is not based on the standard uid attribute.
For example if your dn is employeeNumber=xxxx,dc=domain,dc=com and the
unique member on posix group is based on the employeeNumber. You may not
be able to authenticate to group if authentication is based on uid attribute.
So this will help to match any LDAP structure.

=back
 
=over 4

=item B<UserGroupDN>

The base distinguished name with which to query all LDAP posix group LDAP
that a user belongs to.  By default, the UserGroupDN is empty.

This will set the Apache environment variable REMOTE_GROUP with the require
group and an other variable REMOTE_ALLGROUP with a string containing all
groups of the user separated by a ';'

=back
  
=over 4

=item B<ExtraAttr>

The attribute name that contains a list of LDAP attributes
name separated with a ';' that you may want to be return as
an Apache environment variable. The Apache environment variable
will be named 'REMOTE_' . uc(ExtraAttr).
By default, ExtraAttr is empty.

example:  PerlSetVar ExtraAttr "departmentNumber;employeeNumber"
will return two Apache environment variable:

REMOTE_DEPARTMENTNUMBER and REMOTE_EMPLOYEENUMBER

So you can use them into your perl CGI programs

=back
  
=over 4

=item B<LogoutPattern>

The pattern to match in URI arguments to emulate a logout.
By default, LogoutPattern is set to 'logout'. To disable
this set it to "off".

example:  PerlSetVar LogoutPattern "?logmeoutofSSO"
If this pattern is found into the arguments of the URL this module
will expire the current session if the timestamp given as value of
this argument is upper than the request time and then ask for
authentication again.

WARNING: the value of the argument must be a date in seconde like
the value returned by the perl time function + some little delay.

See LOGOUT.readme file for more information.

=back


=head1 CACHING CONFIGURATION OPTIONS

The following variables are used to enable the caching of authentication

=over 4

=item B<AuthCache>

Enable or disable auth caching. By default, AuthCache is set to off.
To enable caching set it to on.

=back

=over 4

=item B<AuthCacheSuppressPort>

Disable per port caching. By default, AuthCacheSuppressPort is set to off.
To enable single signon on any apache listening port set it to on.

=back

=over 4

=item B<AuthCacheSingleSignon>

Allow single signon on the current apache running domain. Default is off, no SSO.
Set it to on to have SSO on all apache servers running on a single domain.

Note that the B<AuthName> must be the same on each servers. This module only
support AuthType Basic.

The name of the SSO cookie is build as follow:

	Apache-AuthCache-domain_name:server_port-auth_type:auth_name=VALUE

where 'domain_name' is the second level domain name of the server, ex: .samse.fr,
'server_port' is the listening http port or 0 if AuthCacheSuppressPort is set
to off. Here a complete example of a cookie name:

    Apache-AuthCache-.samse.fr:0-Basic:Intranet Authentication

The VALUE of the cookie is build as follow:

	username:passwd:remote_ip:timestamp:remote_group:remote_allgroup:other_name=other_val

the VALUE is base64 encoded, so it may only be used in a secure environment.

'other_name' variables are those defined in the ExtraAttr perl directive and 'other_val'
is the values of the corresponding LDAP attributes retrieved from LDAP directory.

For example if you have defined a perl directive in the httpd.conf as follow:

    PerlSetVar ExtraAttr "departmentNumber;employeeNumber;codeAgence"

the authentication module will retrieve the values of these LDAP attributes from
the LDAP directory and 'other_name' and 'other_val' would be:

    departmentNumber=NNN:employeeNumber=NNN:codeAgence=NNN

Here a complete example of the cookie value:

    gsample:Linux64:192.168.2.34:1076058109:intranet:;intranet;users;internet;:codeAgence=331:mail=guest-sample@samse.fr:cn=SAMPLE Guest:departmentNumber=344:employeeNumber=000111

In this example the ExtraAttr perlvar directive is set as follow:

    PerlSetVar ExtraAttr "employeeNumber;departmentNumber;mail;codeAgence;cn"

Note that in the remote_allgroup part of the cookie the separator of group name
(;) could be present or not at begin and end of the section.

=back

=over 4

=item B<AuthCacheTTL>

Time to live of caching in second. Default is 0, no expiration at all.
Set it to the number second you want the cache expire.

=back


=head1 AUTHORS

Gilles Darold <gilles@darold.net>

=head1 See ALSO

httpd(8), ldap(3), mod_perl(1), slapd(8C)

=head1 COPYRIGHT
 
Copyright (C) 2002-2005, Gilles Darold - Groupe SAMSE
All Rights Reserved.
 
This module is free software; you can redistribute it and/or
modify it under the terms of Perl itself.

=cut

